# -*- coding: utf-8 -*-

from gimmik.base import MatMul


class ISPCMatMul(MatMul):
    platform = 'ispc'
    basemeta = {}

    def _kernel_generators(self, dtype, dsize):
        yield ('cstream', {}, {})



# -*- coding: utf-8 -*-

from gimmik.base import MatMul


class HIPMatMul(MatMul):
    platform = 'hip'
    basemeta = {'block': (128, 1, 1), 'width': 1, 'shared': 0}

    def _kernel_generators(self, dtype, dsize, *, gcn_arch=None, warp_size=64):
        # B loading, C streaming kernel
        yield ('cstream', {}, {})

        # B streaming, C accumulation kernel
        yield ('bstream', {}, {})

        # Four-way m-split B streaming, C accumulation kernel
        ms, bsz, blkx = 4, 24, 64
        args = {'msplit': ms, 'bsz': bsz, 'blockx': blkx}
        meta = {'block': (blkx, ms, 1), 'shared': 2*bsz*blkx*dsize}
        yield ('bstream-msplit', args, meta)

        ## Two-way k-split B loading, C streaming kernel
        ks, csz, blkx = 2, 24, 64
        args = {'ksplit': ks, 'csz': csz, 'blockx': blkx}
        meta = {'block': (blkx, ks, 1), 'shared': (ks - 1)*csz*blkx*dsize}
        yield ('cstream-ksplit', args, meta)

        # At single precision also consider vectorized kernels
        if (dtype == 'float' and
            self.aligne is not None and self.aligne % 2 == 0):
            # Vector B loading, C streaming kernel
            args = {'dtype': 'float2', 'width': 2}
            meta = {'width': 2}
            yield ('cstream', args, meta)

            # Vector four-way m-split B streaming, C accumulation kernel
            ms, bsz, blkx = 4, 16, 64
            args = {'dtype': 'float2', 'width': 2, 'msplit': ms,
                    'bsz': bsz, 'blockx': blkx}
            meta = {'block': (blkx, ms, 1), 'width': 2,
                    'shared': 2*blkx*bsz*2*dsize}
            yield ('bstream-msplit', args, meta)

            # Vector two-way k-split B loading, C streaming kernel
            ks, csz, blkx = 2, 24, 64
            args = {'dtype': 'float2', 'width': 2, 'ksplit': ks,
                    'csz': csz, 'blockx': blkx}
            meta = {'block': (blkx, ks, 1), 'width': 2,
                    'shared': 2*(ks - 1)*csz*blkx*dsize}
            yield ('cstream-ksplit', args, meta)

    def _process_meta(self, meta):
        if self.n is not None:
            div = meta['block'][0]*meta['width']
            meta['grid'] = (-(-self.n // div), 1, 1)



# -*- coding: utf-8 -*-

from gimmik.base import MatMul


class OpenCLMatMul(MatMul):
    platform = 'opencl'
    basemeta = {'local_work_size': None, 'local_mem_size': 0, 'width': 1}

    def _kernel_generators(self, dtype, dsize, *, local_mem_size=None):
        max_local_mem = local_mem_size or 1024**3

        # B loading, C streaming kernel
        yield ('cstream', {}, {})

        # B streaming, C accumulation kernel
        yield ('bstream', {}, {})

        # Four-way m-split B streaming, C accumulation kernel
        ms, bsz, blkx = 4, 16, 64
        args = {'msplit': ms, 'blockx': blkx, 'bsz': bsz}
        meta = {'local_work_size': (blkx, ms),
                'local_mem_size': 2*blkx*bsz*dsize}
        if meta['local_mem_size'] < max_local_mem:
            yield ('bstream-msplit', args, meta)

        # Two-way k-split B loading, C streaming kernel
        ks, csz, blkx = 2, 32, 64
        args = {'ksplit': ks, 'csz': csz, 'blockx': blkx}
        meta = {'local_work_size': (blkx, ks),
                'local_mem_size': (ks - 1)*csz*blkx*dsize}
        if meta['local_mem_size'] < max_local_mem:
            yield ('cstream-ksplit', args, meta)

        # At single precision also consider vectorized kernels
        if (dtype == 'float' and
            self.aligne is not None and self.aligne % 2 == 0):
            # Vector B loading, C streaming kernel
            args = {'dtype': 'float2', 'width': 2}
            meta = {'width': 2}
            yield ('cstream', args, meta)

            # Vector four-way m-split B streaming, C accumulation kernel
            ms, bsz, blkx = 4, 16, 64
            args = {'dtype': 'float2', 'width': 2, 'msplit': ms,
                    'blockx': blkx, 'bsz': bsz}
            meta = {'local_work_size': (blkx, ms),
                    'local_mem_size': 2*blkx*bsz*dsize, 'width': 2}
            if meta['local_mem_size'] < max_local_mem:
                yield ('bstream-msplit', args, meta)

    def _process_meta(self, meta):
        if self.n is not None:
            lws, width = meta['local_work_size'], meta['width']
            if lws is not None:
                meta['global_work_size'] = (-(-self.n // width), lws[1])
            else:
                meta['global_work_size'] = (-(-self.n // width),)



# -*- coding: utf-8 -*-

import itertools as it
import pkgutil
import re

from mako.lookup import TemplateLookup
from mako.template import Template
import numpy as np


class _PlatformTemplateLookup(TemplateLookup):
    def __init__(self, platform):
        self.platform = platform

    def adjust_uri(self, uri, relto):
        return uri

    def get_template(self, name):
        platform = self.platform
        src = pkgutil.get_data(__name__, f'kernels/{platform}/{name}.mako')

        return Template(src, lookup=self)


def _dot(bfn, row, maxsplit=1):
    nzixs, = np.nonzero(row)

    if not nzixs.size:
        return '0.0'

    nsplit = max(min(maxsplit, nzixs.size // 3), 1)
    snzixs = np.array_split(nzixs, nsplit)

    frags = [' + '.join(f'{row[i]}*{bfn(i)}' for i in ix) for ix in snzixs]
    return ' + '.join(f'({f})' for f in frags)


def _partition(mat, into, by):
    if by == 'rows':
        return [list(range(i, len(mat), into)) for i in range(into)]
    elif by == 'cols':
        return [list(range(i, len(mat.T), into)) for i in range(into)]
    else:
        raise ValueError('Invalid partition by')


def _chunk(l, chunksz):
    l, n = iter(l), len(l)
    nchunks = -(-n // chunksz)

    return [list(it.islice(l, chunksz)) for i in range(nchunks)]


class MatMul:
    platform = None

    def __init__(self, A, beta=0.0, aligne=None, n=None, ldb=None, ldc=None):
        self.A = A
        self.beta = beta
        self.aligne = aligne

        if n is None and ldb is None and ldc is None:
            self.n = self.ldb = self.ldc = None
        elif n is not None and ldb is not None and ldc is not None:
            if aligne is not None and (ldb % aligne or ldc % aligne):
                raise ValueError('ldb/ldc not compatible with aligne')

            self.n, self.ldb, self.ldc = n, ldb, ldc
        else:
            raise ValueError('Must provide all of (n, ldb, ldc) or none')

        # Check the matrix has a non-zero
        if not A.any():
            raise ValueError('A can not be empty')

        # Extract the shape of A
        self.m, self.k = m, k = A.shape

        # Determine the index of the first and last non-zero in each row of A
        self.afix = (A != 0).argmax(axis=1)
        self.alix = k - 1 - (A != 0)[:, ::-1].argmax(axis=1)

        # Mark rows of A which are all zero
        self.afix = np.where(np.any(A != 0, axis=1), self.afix, -1)
        self.alix = np.where(np.any(A != 0, axis=1), self.alix, -1)
        self.has_zero_rows = np.any(self.afix == -1)

        # Determine which entries of B partake in the multiplication
        self.bix = np.nonzero(np.any(A != 0, axis=0))[0]
        self.bix = {kx: k for k, kx in enumerate(self.bix)}

    def kernels(self, dtype, kname='gimmik_mm', **kwargs):
        basemeta = self.basemeta

        # Process the data type
        dtype = np.dtype(dtype).type
        if dtype == np.float32:
            dtype, dsize = 'float', 4
        elif dtype == np.float64:
            dtype, dsize = 'double', 8
        else:
            raise ValueError('Invalid floating point data type')

        # Common template arguments
        baseargs = {
            'dtype': dtype, 'kname': kname,
            'A': self.A, 'beta': self.beta, 'width': 1,
            'm': self.m, 'n': self.n, 'k': self.k,
            'ldb': self.ldb, 'ldc': self.ldc,
            'afix': self.afix, 'alix': self.alix, 'bix': self.bix,
            'dot': _dot, 'partition': _partition, 'chunk': _chunk
        }

        # Incrementally generate and render the kernels
        gen = self._kernel_generators(dtype, dsize, **kwargs)
        try:
            resp = None
            while True:
                # Generate the next kernel in the sequence
                name, exargs, exmeta = gen.send(resp)

                # Merge in the base arguments and metadata
                args = baseargs | exargs
                meta = basemeta | exmeta

                # Render the kernel template
                src = self._render_kernel(dtype, name, args)

                # Post-process the metadata
                meta['tplname'] = name
                self._process_meta(meta)

                # Yield the source and metadata and await a response
                resp = yield (src, meta)
        except StopIteration:
            pass

    def _process_meta(self, meta):
        pass

    def _render_kernel(self, dtype, tplname, tplargs):
        tpl = _PlatformTemplateLookup(self.platform).get_template(tplname)
        src = tpl.render(**tplargs)

        # At single precision suffix all floating point constants by 'f'
        if dtype == 'float':
            src = re.sub(r'(?=\d*[.eE])(?=\.?\d)\d*\.?\d*(?:[eE][+-]?\d+)?',
                         r'\g<0>f', src)

        # Cleanup
        src = re.sub(r'^\w+\n$', '', src.strip())
        src = re.sub(r'\n\n+', r'\n\n', src) + '\n'
        src = re.sub(r'\w+$', '', src)
        return src



# -*- coding: utf-8 -*-

from gimmik._version import __version__
from gimmik.c import CMatMul
from gimmik.copenmp import COpenMPMatMul
from gimmik.cuda import CUDAMatMul
from gimmik.ispc import ISPCMatMul
from gimmik.hip import HIPMatMul
from gimmik.metal import MetalMatMul
from gimmik.opencl import OpenCLMatMul


def generate_mm(mat, dtype, platform, alpha=1.0, beta=0.0, funcn='gimmik_mm',
                n=None, ldb=None, ldc=None):
    import warnings

    warnings.warn('generate_mm is deprecated, use MatMul', DeprecationWarning)

    platmap = {
        'c': CMatMul,
        'c-omp': COpenMPMatMul,
        'cuda': CUDAMatMul,
        'ispc': ISPCMatMul,
        'hip': HIPMatMul,
        'opencl': OpenCLMatMul
    }

    mm = platmap[platform](alpha*mat, beta, None, n, ldb, ldc)
    return next(mm.kernels(dtype, kname=funcn))[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
from setuptools import setup
import sys


# Python version
if sys.version_info[:2] < (3, 9):
    print('GiMMiK requires Python 3.9 or newer')
    sys.exit(-1)

# GiMMiK version
vfile = open('gimmik/_version.py').read()
vsrch = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", vfile, re.M)

if vsrch:
    version = vsrch.group(1)
else:
    print('Unable to find a version string in gimmik/_version.py')

# Data
package_data = {
    'gimmik': ['kernels/*/*.mako'],
}

# Hard dependencies
install_requires = [
    'mako',
    'numpy >= 1.7'
]

# Info
classifiers = [
    'License :: OSI Approved :: BSD License',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Topic :: Scientific/Engineering'
]

# Long Description
long_description = '''GiMMiK is a Python based kernel generator for
matrix multiplication kernels for various accelerator platforms.  For
small operator matrices the generated kernels are capable of
outperfoming the state-of-the-art general matrix multiplication
routines such as cuBLAS GEMM or clBLAS GEMM.  GiMMiK was originally
developed as part of Bartosz Wozniak's master's thesis in the
Department of Computing at Imperial College London and is currently
maintained by Freddie Witherden.'''

# Keywords
keywords = ['Matrix Multiplication', 'ISPC', 'GPU', 'CUDA', 'HIP', 'Metal',
            'OpenCL']

setup(name='gimmik',
      version=version,

      # Packages
      packages=['gimmik'],
      package_data=package_data,
      install_requires=install_requires,

      # Metadata
      description='Generator of Matrix Multiplication Kernels',
      long_description=long_description,
      maintainer='Freddie Witherden',
      maintainer_email='freddie@witherden.org',
      url='https://github.com/vincentlab/GiMMiK',
      license='BSD',
      keywords=keywords,
      classifiers=classifiers)



# -*- coding: utf-8 -*-

from gimmik.base import MatMul


class MetalMatMul(MatMul):
    platform = 'metal'
    basemeta = {'threadgroup': (128, 1, 1), 'threadgroup_mem_size': 0,
                'width': 1}

    def _kernel_generators(self, dtype, dsize):
        # B loading, C streaming kernel
        yield ('cstream', {}, {})

        # B streaming, C accumulation kernel
        yield ('bstream', {}, {})

        # Four-way m-split B streaming, C accumulation kernel
        ms, bsz, blkx = 4, 16, 32
        args = {'msplit': ms, 'blockx': blkx, 'bsz': bsz}
        meta = {'threadgroup': (blkx, ms, 1),
                'threadgroup_mem_size': 2*blkx*bsz*dsize}
        yield ('bstream-msplit', args, meta)

        # Four-way m-split B streaming, C accumulation kernel
        ms, bsz, blkx = 4, 20, 32
        args = {'msplit': ms, 'blockx': blkx, 'bsz': bsz}
        meta = {'threadgroup': (blkx, ms, 1),
                'threadgroup_mem_size': 2*blkx*bsz*dsize}
        yield ('bstream-msplit', args, meta)

        # Two-way k-split B loading, C streaming kernel
        ks, csz, blkx = 2, 20, 32
        args = {'ksplit': ks, 'csz': csz, 'blockx': blkx}
        meta = {'threadgroup': (blkx, ks, 1),
                'threadgroup_mem_size': (ks - 1)*csz*blkx*dsize}
        yield ('cstream-ksplit', args, meta)

        if self.aligne is not None and self.aligne % 2 == 0:
            # Vector B loading, C streaming kernel
            args = {'dtype': 'float2', 'width': 2}
            meta = {'width': 2}
            yield ('cstream', args, meta)

            # Vector B streaming, C accumulation kernel
            yield ('bstream', args, meta)

            # Vector four-way m-split B streaming, C accumulation kernel
            ms, bsz, blkx = 4, 16, 32
            args = {'dtype': 'float2', 'width': 2, 'msplit': ms,
                    'blockx': blkx, 'bsz': bsz}
            meta = {'threadgroup': (blkx, ms, 1),
                    'threadgroup_mem_size': 2*blkx*bsz*dsize, 'width': 2}
            yield ('bstream-msplit', args, meta)

    def _process_meta(self, meta):
        if self.n is not None:
            tg = meta['threadgroup']
            meta['grid'] = (-(-self.n // meta['width']), tg[1], 1)
